//Name:		ChecksumDataBuffer.h
//Author:	Brian Tietz
//Description:
//	A buffer for writing data to for network transmission.  The buffer grows as needed to hold the
//	data as it's added, and can return a checksum.  The recommended checksum method is to use a 48-bit
//	checksum by sending both GetChecksum16() and GetWrappingChecksum().  GetChecksum just adds each byte
//	of the data buffer to a uint16 or uint32, depending on the function version.  GetWrappingChecksum
//	generates the sum for all bytes in the data buffer of (byte_value<<(pos%25)) in a uint32.  On the
//	receiving end, the static versions of the checksum functions can be used to generate a checksum
//	on read data by the same algorithm.
//
//Usage:
//	For sending:
//	  - Add data with the various AddXXX() calls
//	  - Replace data that was added as a placeholder, such as length info near the beginning of the
//		buffer
//	  - Generate a 16-bit checksum and a 32-bit wrapping checksum, add them to the buffer
//	  - Write the whole buffer
//	For receiving:
//	  - Call ExpandBufferForAdd with your initial read length
//	  - Call your external read function on ExternalAddDataAddress()
//	  - Call ExternalAddDataCompleted to let ChecksumDataBuffer know how much data was really added
//	  - Read data with various ReadXXX() calls
//	  - Call ExpandBufferForAdd(), ExternalAddDataAddress(), ExternalAddDataCompleted() as needed
//		to read more data based on what was read previously, such as length info
//	  - Generate a 16-bit checksum and a 32-bit wrapping checksum (you should not have read them into
//		the buffer yet)
//	  - Call ExpandBufferForAdd(), ExternalAddDataAddress(), ExternalAddDataCompleted() as needed
//		to read the 16-bit checksum and 32-bit wrapping checksum into the buffer
//	  - Call ReadUint16At() and ReadUint32At() to read the checksums, see if they match what was
//		generated from the buffer data
//	  - If the checksums match, read the rest of the data with various ReadXXX() calls
//
//Conventions:
//	Global constants (declared with const) and #defines - begin with "c_" followed by lowercase
//		words separated by underscores.
//		(E.G., #define c_my_constant 5).
//		(E.G., const int c_my_constant = 5;).
//	Global variables - begin with "g_" followed by lowercase words separated by underscores.
//		(E.G., int g_my_global;).
//	New data types (classes, structs, typedefs, etc.) - begin with an uppercase letter followed by
//		lowercase words separated by uppercase letters.  Enumerated constants contain a prefix
//		associating them with a particular enumerated set.
//		(E.G., typedef int MyTypedef;).
//		(E.G., enum MyEnumConst {c_mec_one, c_mec_two};)
//	Private member variables - begin with "m_" followed by lowercase words separated by underscores.
//		(E.G., int m_my_member;).
//	Public or friend-accessible member variables - all lowercase words separated by underscores.
//		(E.G., int public_member;).
//	Argument and local variables - begin with a lowercase letter followed by
//		lowercase words separated by underscores.  If the name is already taken by a public member
//		variable, prefix with a_ or l_
//		(E.G., int my_local; int a_my_arg, int l_my_local).
//	Functions (member or global) - begin with an uppercase letter followed by lowercase words
//		separated by uppercase letters.
//		(E.G., void MyFunction(void);).

#ifndef _CHECKSUM_DATA_BUFFER_H_
#define _CHECKSUM_DATA_BUFFER_H_


//******************************************************************************************************
//**** System header files
//******************************************************************************************************
#include <SupportDefs.h>


//******************************************************************************************************
//**** ChecksumDataBuffer
//******************************************************************************************************
class ChecksumDataBuffer
{
	public:
		ChecksumDataBuffer(uint32 growth_increment, bool use_network_order);
		~ChecksumDataBuffer();

		inline void AddInt64(int64 data) {AddNumber(&data,8);}
		inline void AddInt32(int32 data) {AddNumber(&data,4);}
		inline void AddInt16(int16 data) {AddNumber(&data,2);}
		inline void AddInt8(int8 data) {AddNumber(&data,1);}
		inline void AddUint64(uint64 data) {AddNumber(&data,8);}
		inline void AddUint32(uint32 data) {AddNumber(&data,4);}
		inline void AddUint16(uint16 data) {AddNumber(&data,2);}
		inline void AddUint8(uint8 data) {AddNumber(&data,1);}
		void AddNumber(void* data, uint8 size);
		void AddData(const void* data, uint32 length);
			//This is safe to call with a NULL pointer if length is 0

		inline void ReplaceInt64(int64 data, int32 at_pos) {ReplaceNumber(&data,8,at_pos);}
		inline void ReplaceInt32(int32 data, int32 at_pos) {ReplaceNumber(&data,4,at_pos);}
		inline void ReplaceInt16(int16 data, int32 at_pos) {ReplaceNumber(&data,2,at_pos);}
		inline void ReplaceInt8(int8 data, int32 at_pos) {ReplaceNumber(&data,1,at_pos);}
		inline void ReplaceUint64(uint64 data, int32 at_pos) {ReplaceNumber(&data,8,at_pos);}
		inline void ReplaceUint32(uint32 data, int32 at_pos) {ReplaceNumber(&data,4,at_pos);}
		inline void ReplaceUint16(uint16 data, int32 at_pos) {ReplaceNumber(&data,2,at_pos);}
		inline void ReplaceUint8(uint8 data, int32 at_pos) {ReplaceNumber(&data,1,at_pos);}
		void ReplaceNumber(void* data, uint8 size, uint32 at_pos);
		void ReplaceData(const void* data, uint32 length, uint32 at_pos);
			//This is safe to call with a NULL pointer if length is 0

		inline uint8* Data() const {return m_data;}
		inline uint32 Size() const {return m_buffer_pos;}

		void ExpandBufferForAdd(uint32 length);
			//Makes room in the buffer for an add to occur from an external source (typically, reading
			//from a socket)
		inline uint8* ExternalAddDataAddress() const {return &m_data[m_buffer_pos];}
			//The address where an external add should , provided ExpandBufferForAdd has been called.
		inline void ExternalAddDataCompleted(uint32 length) {m_buffer_pos += length;}
			//Updates the data area size to include data that was added from an external source.

		inline int64 ReadInt64() {int64 value; ReadNumber(&value,8); return value;}
		inline int32 ReadInt32() {int32 value; ReadNumber(&value,4); return value;}
		inline int16 ReadInt16() {int16 value; ReadNumber(&value,2); return value;}
		inline int8  ReadInt8() {int8 value; ReadNumber(&value,1); return value;}
		inline uint64 ReadUint64() {uint64 value; ReadNumber(&value,8); return value;}
		inline uint32 ReadUint32() {uint32 value; ReadNumber(&value,4); return value;}
		inline uint16 ReadUint16() {uint16 value; ReadNumber(&value,2); return value;}
		inline uint8  ReadUint8() {uint8 value; ReadNumber(&value,1); return value;}
		inline int64 ReadInt64At(uint32 at_pos) const {int64 value; ReadNumberAt(&value,8,at_pos); return value;}
		inline int32 ReadInt32At(uint32 at_pos) const {int32 value; ReadNumberAt(&value,4,at_pos); return value;}
		inline int16 ReadInt16At(uint32 at_pos) const {int16 value; ReadNumberAt(&value,2,at_pos); return value;}
		inline int8  ReadInt8At(uint32 at_pos) const {int8 value; ReadNumberAt(&value,1,at_pos); return value;}
		inline uint64 ReadUint64At(uint32 at_pos) const {uint64 value; ReadNumberAt(&value,8,at_pos); return value;}
		inline uint32 ReadUint32At(uint32 at_pos) const {uint32 value; ReadNumberAt(&value,4,at_pos); return value;}
		inline uint16 ReadUint16At(uint32 at_pos) const {uint16 value; ReadNumberAt(&value,2,at_pos); return value;}
		inline uint8  ReadUint8At(uint32 at_pos) const {uint8 value; ReadNumberAt(&value,1,at_pos); return value;}
		void ReadNumber(void* data, uint8 size);
		void ReadNumberAt(void* data, uint8 size, uint32 at_pos)  const;
		inline const uint8* ReadData(uint32 length)
			{uint8* read_pos = &m_data[m_read_pos]; m_read_pos+=length; return ((length>0)?read_pos:NULL);}

		inline uint16 GetChecksum16() const {return GetChecksum16(m_data,m_buffer_pos);}
		inline uint32 GetChecksum32() const {return GetChecksum32(m_data,m_buffer_pos);}
		inline uint32 GetWrappingChecksum() const {return GetWrappingChecksum(m_data,m_buffer_pos);}
		inline void Flush() {m_buffer_pos = 0; m_read_pos = 0;}

		static uint16 GetChecksum16(uint8* data, uint32 length);
		static uint32 GetChecksum32(uint8* data, uint32 length);
		static uint32 GetWrappingChecksum(uint8* data, uint32 length);
		static uint32 AddMoreWrappingChecksum(uint8* data, uint32 length, uint32 prev_data_length,
			uint32 prev_checksum);
			//To send in chunks and calculate a checksum on all chunks together (checksum on a buffer
			//too big to reasonably fit in memory at once, for example)

	private:
		uint8* m_data;
		uint32 m_growth_increment;
		uint32 m_buffer_size;
		uint32 m_buffer_pos;
		uint32 m_read_pos;
		bool m_use_network_order;
};


#endif //_CHECKSUM_DATA_BUFFER_H_
